searchNodes=[{"doc":"Carbonite implements the Change-Data-Capture pattern on top of PostgreSQL using database triggers. It keeps a central changes table where all mutations of participating tables are recorded. Each changes row is associated to a single row in the transactions table using PostgreSQL's internal transaction id as the foreign key. This leads to the following interesting properties: All changes created within a database transaction automatically and implicitly belong to the same record in the transactions table, even if they're created separatedly and agnostic of each other in the application logic. This gives the developer a &quot;natural&quot; way to group related changes into events (more on events later). As the changes table is associated to the transactions table via a non-nullable foreign key constraint, the entry in the transactions table must be created before any changes . Attempting to modify a versioned table without prior insertion into the transactions table will result in an error. The transactions table carries transactional metadata which can be set by the developer on creation. Consequently, much of Carbonite's logic lives in database functions and triggers. To get started, we need to create a migration using Ecto. ℹ️ Trigger vs. Write-Ahead-Log Existing solutions for CDC on top of a PostgreSQL database (e.g., Debezium ) often tail the &quot;Write-Ahead-Log&quot; instead of using database logic &amp; triggers to create change records. While this is likely more performant than using triggers, it makes it difficult to correlate changes on a transaction level as Carbonite does, and has different consistency guarantees. Carbonite makes this trade-off in favour of simplicity and universality. You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before. Installing the schema &amp; triggers The following migration installs Carbonite into its &quot;default prefix&quot;, a PostgreSQL schema aptly called carbonite_default , and installs the change capture trigger for an exemplary table called rabbits (in the public schema). In a real-world scenario, you will most likely want to install the trigger for a set of tables and optionally split the audit trail into multiple partitions. See Carbonite.Migrations for more information on migrations. mix ecto.gen.migration InstallCarbonite # priv/repo/migrations/20210704201534_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do use Ecto.Migration def up do # Creates carbonite_default schema and tables. Carbonite.Migrations . install_schema ( ) # For each table that you want to capture changes of, you need to install the trigger. Carbonite.Migrations . install_trigger ( :rabbits ) end def down do # Remove all triggers before dropping the schema. Carbonite.Migrations . drop_trigger ( :rabbits ) # Drop the schema &amp; tables. Carbonite.Migrations . drop_schema ( ) end end Primary Key Columns To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the table_pk column of the changes table. The table keeps an index on this column together with the table prefix and name. By default, Carbonite will try to copy the :id column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the primary_key_columns option of Carbonite.Migrations.install_trigger/2 and Carbonite.Migrations.configure_trigger/2 . # Disable PK copying Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ ] ) # Different name Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;identifier&quot; ] ) # Composite PK Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;house&quot; , &quot;apartment_no&quot; ] ) Since the changes table keeps versions of a multitude of different source tables, primary keys are first cast to string (the table_pk column has type VARCHAR[] ). For composite primary keys, set the primary_key_columns option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array. Excluded and Filtered Columns In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the excluded_columns option. Excluded columns will not appear in the captured data. If an UPDATE on a table solely touches excluded columns, the entire UPDATE will not be recorded. Carbonite.Migrations . install_trigger ( :rabbits , excluded_columns : [ &quot;age&quot; ] ) If you still want to capture changes to a column (in the changed field), but don't need the exact data, you can make it a &quot;filtered&quot; column. These columns appear as &quot;[FILTERED]&quot; in the data field. ```elixir Carbonite.Migrations.install_trigger(:rabbits, filtered_columns: [&quot;age&quot;]) ``` If you forgot to exclude a column, you can reconfigure a trigger for a particular table using configure trigger/2 : ```elixir # in another migration Carbonite.Migrations.configure_trigger(:rabbits, excluded_columns: [&quot;age&quot;]) ``` #### Partitioning the Audit Trail Carbonite can install its tables into multiple database schemas using the prefix option. You can use this feature to &quot;partition&quot; your captured data. ```elixir Carbonite.Migrations.install_schema(carbonite_prefix: &quot;carbonite_lagomorpha&quot;) Carbonite.Migrations.install_trigger(:rabbits, carbonite_prefix: &quot;carbonite_lagomorpha&quot;) ``` If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage. ## Inserting a Transaction In your application logic, before modifying a versioned table like rabbits , you need to first create a Carbonite.Transaction record. ### With Ecto.Multi The easiest way to do so is using Carbonite.Multi.insert_transaction/2 within an Ecto.Multi operation: ```elixir Ecto.Multi.new() |&gt; Carbonite.Multi.insert_transaction(%{meta: %{type: &quot;rabbit_inserted&quot;}}) |&gt; Ecto.Multi.insert(:rabbit, &amp;MyApp.Rabbit.create_changeset(&amp;1.params)) |&gt; MyApp.Repo.transaction() ``` As you can see, the Carbonite.Transaction is a great place to store metadata for the operation. A &quot;transaction type&quot; would be an obvious choice to categorize the transactions. A user_id would be a good candidate for an transaction log, as well. ### Building a changeset for manual insertion If you don't have the luxury of an Ecto.Multi , you can create a changeset for a Carbonite.Transaction using Carbonite.Transaction.changeset/1 : ```elixir MyApp.Repo.transaction(fn -&gt; %{meta: %{type: &quot;rabbit_inserted&quot;}} |&gt; Carbonite.Transaction.changeset() |&gt; MyApp.Repo.insert!() # ... end) ``` ### Setting metadata outside of the Transaction In case you do not have access to metadata you want to persist in the Carbonite.Transaction at the code site where you create it, you can use Carbonite.Transaction.put_meta/2` to store metadata in the _process dictionary . This metadata is merged into the metadata given to Carbonite.Multi.insert_transaction/2 . elixir # e.g., in a controller or plug Carbonite.Transaction.put_meta(:user_id, ...) ## Retrieving data Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. Carbonite.Query offers a small suite of helper functions that make it easier to query the database for Transaction and Change records. ### Fetching Changes The Carbonite.Query.changes/2 function constructs an Ecto.Query from a schema struct, loading all changes stored for the given source record. elixir %MyApp.Rabbit{id: 1} |&gt; Carbonite.Query.changes() |&gt; MyApp.Repo.all() ## Testing / Bypassing Carbonite One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing Carbonite.Transaction beforehand (due to the foreign key constraint between changes and transactions ). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. To bypass the capture trigger, Carbonite's trigger configuration provides a toggle mechanism consisting of two fields: mode and override_transaction_id . The former you set while installing the trigger on a table in a migration, while the latter allows to &quot;override&quot; whatever has been set at runtime, and only for the current transaction. If you are using Ecto's SQL sandbox for running transactional tests, this means the override is going to be active until the end of the test case. As a result, you have two options: 1. Leave the mode at the default value of :capture and turn off capturing as needed by switching to &quot;override mode&quot;. This means for every test case where you do not care about change capturing, you explicitly disable the trigger before any database calls; for instance, in a ExUnit setup block. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a Carbonite.Transaction . It is, however, still pretty expensive at ~1 additional SQL call per test case. 2. Set the mode to :ignore on all triggers in your :test environment and instead selectively turn on capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well. The following code snippet illustrates the second approach: elixir # config/config.exs config :my_app, carbonite_mode: :capture # config/test.exs config :my_app, carbonite_mode: :ignore # priv/repo/migrations/000000000000_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do @mode Application.compile_env!(:my_app, :carbonite_mode) def up do # ... Carbonite.Migrations.install_trigger(:rabbits, mode: @mode) end end # test/support/carbonite_helpers.exs defmodule MyApp.CarboniteHelpers do def carbonite_override_mode(_) do Ecto.Multi.new() |&gt; Carbonite.Multi.override_mode() |&gt; MyApp.Repo.transaction() :ok end end # test/some_test.exs def SomeTest do use MyApp.DataCase import MyApp.CarboniteHelpers setup [:carbonite_override_mode] end","ref":"Carbonite.html","title":"Carbonite","type":"module"},{"doc":"Returns the default audit trail prefix.","ref":"Carbonite.html#default_prefix/0","title":"Carbonite.default_prefix/0","type":"function"},{"doc":"A Carbonite.Change records a mutation on a database table. INSERT statements lead to a Change where the data field contains the inserted row as a JSON object while the changed field is an empty list. UPDATE statements contain the updated record in data while the changed field is a list of attributes that have changed. DELETE statements have the delete data in data while changed is again an empty list.","ref":"Carbonite.Change.html","title":"Carbonite.Change","type":"module"},{"doc":"","ref":"Carbonite.Change.html#t:t/0","title":"Carbonite.Change.t/0","type":"type"},{"doc":"Functions to setup Carbonite audit trails in your migrations.","ref":"Carbonite.Migrations.html","title":"Carbonite.Migrations","type":"module"},{"doc":"Alters a triggers configuration for a given table. You need to specify all desired options each time you call this function, as the previous configuration will be replaced. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot; primary_key_columns is a list of columns that form the primary key of the table ( defaults to ` [ &quot;id&quot; ] ` , set to ` [ ] ` to disable ) excluded_columns is a list of columns to exclude from change captures filtered_columns is a list of columns that appear as '[FILTERED]' in the data mode is either :capture or :ignore and defines the default behaviour of the trigger","ref":"Carbonite.Migrations.html#configure_trigger/2","title":"Carbonite.Migrations.configure_trigger/2","type":"function"},{"doc":"Removes a Carbonite audit trail from the database. Options prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_schema/1","title":"Carbonite.Migrations.drop_schema/1","type":"function"},{"doc":"Removes a change capture trigger from a table. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#drop_trigger/2","title":"Carbonite.Migrations.drop_trigger/2","type":"function"},{"doc":"Installs a Carbonite audit trail. Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Migrations.html#install_schema/1","title":"Carbonite.Migrations.install_schema/1","type":"function"},{"doc":"Installs a change capture trigger on a table. Options table_prefix is the name of the schema the table lives in carbonite_prefix is the schema of the audit trail, defaults to &quot;carbonite_default&quot; primary_key_columns is a list of columns that form the primary key of the table ( defaults to ` [ &quot;id&quot; ] ` , set to ` [ ] ` to disable ) excluded_columns is a list of columns to exclude from change captures filtered_columns is a list of columns that appear as '[FILTERED]' in the data","ref":"Carbonite.Migrations.html#install_trigger/2","title":"Carbonite.Migrations.install_trigger/2","type":"function"},{"doc":"","ref":"Carbonite.Migrations.html#t:column_name/0","title":"Carbonite.Migrations.column_name/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:prefix/0","title":"Carbonite.Migrations.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:schema_option/0","title":"Carbonite.Migrations.schema_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:table_name/0","title":"Carbonite.Migrations.table_name/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:trigger_config_option/0","title":"Carbonite.Migrations.trigger_config_option/0","type":"type"},{"doc":"","ref":"Carbonite.Migrations.html#t:trigger_option/0","title":"Carbonite.Migrations.trigger_option/0","type":"type"},{"doc":"This module provides functions for dealing with audit trails in the context of Ecto.Multi.","ref":"Carbonite.Multi.html","title":"Carbonite.Multi","type":"module"},{"doc":"Adds an insert operation for a Carbonite.Transaction to an Ecto.Multi . Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; params map of params for the Carbonite.Transaction (e.g., :meta )","ref":"Carbonite.Multi.html#insert_transaction/3","title":"Carbonite.Multi.insert_transaction/3","type":"function"},{"doc":"Sets the current transaction to &quot;override mode&quot; for all tables in a translation log.","ref":"Carbonite.Multi.html#override_mode/2","title":"Carbonite.Multi.override_mode/2","type":"function"},{"doc":"","ref":"Carbonite.Multi.html#t:insert_transaction_option/0","title":"Carbonite.Multi.insert_transaction_option/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:override_mode_option/0","title":"Carbonite.Multi.override_mode_option/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:params/0","title":"Carbonite.Multi.params/0","type":"type"},{"doc":"","ref":"Carbonite.Multi.html#t:prefix/0","title":"Carbonite.Multi.prefix/0","type":"type"},{"doc":"Implements the outbox pattern to process (and evict) recorded transactions in the database, in order of insertion.","ref":"Carbonite.Outbox.html","title":"Carbonite.Outbox","type":"module"},{"doc":"Builds an Ecto.Multi that can be used to load Carbonite.Transaction records from database (in order of insertion) and process them one-by-one through a function. Processed items are marked as processed and can be purged. Options batch_size is the size of records to load in one chunk, defaults to 20 min_age is the minimum age of a record, defaults to 300 seconds (see below) prefix is the Carbonite schema, defaults to &quot;carbonite_default&quot; Long running transactions &amp; insertion order A warning: As Carbonite.Transaction records are inserted at the beginning of a database transaction, their inserted_at is already a bit in the past when they become visible to other connections, e.g. to your processing job. This means that in case of long running transactions, Carbonite.Transaction records with more recent inserted_at values might be processed before older ones, and hence the eventual total order of inserted_at in the processed records can not be guaranteed. To mitigate this issue, Carbonite will by default exclude records younger than min_age seconds from processing. Tweak this setting if you have even longer transactions.","ref":"Carbonite.Outbox.html#process/2","title":"Carbonite.Outbox.process/2","type":"function"},{"doc":"Builds an Ecto.Multi that can be used to delete Carbonite.Transaction records and their associated Carbonite.Change rows from the database once they have been successfully processed using process/2 . Options prefix is the Carbonite schema, defaults to &quot;carbonite_default&quot;","ref":"Carbonite.Outbox.html#purge/1","title":"Carbonite.Outbox.purge/1","type":"function"},{"doc":"","ref":"Carbonite.Outbox.html#t:prefix/0","title":"Carbonite.Outbox.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:process_fun/0","title":"Carbonite.Outbox.process_fun/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:process_option/0","title":"Carbonite.Outbox.process_option/0","type":"type"},{"doc":"","ref":"Carbonite.Outbox.html#t:purge_option/0","title":"Carbonite.Outbox.purge_option/0","type":"type"},{"doc":"This module provides query functions for retrieving audit trails from the database.","ref":"Carbonite.Query.html","title":"Carbonite.Query","type":"module"},{"doc":"Returns an Ecto.Query that can be used to select changes for a single record. Given an Ecto.Schema struct, this function builds a query that fetches all changes recorded for it from the database, ordered ascending by their ID (i.e., roughly by insertion date descending). Example % MyApp.Rabbit { id : 1 } |&gt; Carbonite.Query . changes ( ) |&gt; MyApp.Repo . all ( ) Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload can be used to preload the transaction","ref":"Carbonite.Query.html#changes/2","title":"Carbonite.Query.changes/2","type":"function"},{"doc":"Returns an Ecto.Query that can be used to select or delete the &quot;current&quot; transaction. This function is useful when your tests run in a database transaction using Ecto's SQL sandbox. Examples Asserting on the current transaction When you insert your Carbonite.Transaction record somewhere inside your domain logic, you do not wish to return it to the caller only to be able to assert on its attributes in tests. This is how you can assert on the current transaction: Carbonite.Query . current_transaction ( ) |&gt; MyApp.Repo . all ( ) # Preload changes Carbonite.Query . current_transaction ( preload : :changes ) |&gt; MyApp.Repo . all ( ) # Same Carbonite.Query . current_transaction ( preload : true ) |&gt; MyApp.Repo . all ( ) Erasing the current transaction Sometimes your test code may run a particular function twice (in multiple transactions if it wasn't for the SQL sandbox), in which case you may need to delete the Carbonite.Transaction inserted first in between. # This deletes the transaction as well as any associated change (per FK constraint) Carbonite.Query . current_transaction ( ) |&gt; MyApp.Repo . delete_all ( ) Options carbonite_prefix defines the audit trail's schema, defaults to &quot;carbonite_default&quot; preload can be used to preload the changes","ref":"Carbonite.Query.html#current_transaction/1","title":"Carbonite.Query.current_transaction/1","type":"function"},{"doc":"","ref":"Carbonite.Query.html#t:changes_option/0","title":"Carbonite.Query.changes_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:current_transaction_option/0","title":"Carbonite.Query.current_transaction_option/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:prefix/0","title":"Carbonite.Query.prefix/0","type":"type"},{"doc":"","ref":"Carbonite.Query.html#t:preload/0","title":"Carbonite.Query.preload/0","type":"type"},{"doc":"A Carbonite.Transaction is the binding link between change records of tables. As such, it contains a set of optional metadata that describes the transaction.","ref":"Carbonite.Transaction.html","title":"Carbonite.Transaction","type":"module"},{"doc":"Builds a changeset for a new Carbonite.Transaction . The :meta map from the params will be merged with the metadata currently stored in the process dictionary.","ref":"Carbonite.Transaction.html#changeset/1","title":"Carbonite.Transaction.changeset/1","type":"function"},{"doc":"Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#current_meta/0","title":"Carbonite.Transaction.current_meta/0","type":"function"},{"doc":"Stores a piece of metadata in the process dictionary. This can be useful in situations where you want to record a value at a system boundary (say, the user's account_id ) without having to pass it through to the database transaction. Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#put_meta/2","title":"Carbonite.Transaction.put_meta/2","type":"function"},{"doc":"","ref":"Carbonite.Transaction.html#t:meta/0","title":"Carbonite.Transaction.meta/0","type":"type"},{"doc":"","ref":"Carbonite.Transaction.html#t:t/0","title":"Carbonite.Transaction.t/0","type":"type"},{"doc":"A Carbonite.Trigger stores per table configuration for the change capture trigger.","ref":"Carbonite.Trigger.html","title":"Carbonite.Trigger","type":"module"},{"doc":"","ref":"Carbonite.Trigger.html#t:mode/0","title":"Carbonite.Trigger.mode/0","type":"type"},{"doc":"","ref":"Carbonite.Trigger.html#t:t/0","title":"Carbonite.Trigger.t/0","type":"type"},{"doc":"Carbonite makes it easy to create audit trails for tables in a PostgreSQL database and integrate them into an Elixir application. Carbonite implements the Change-Data-Capture pattern on top of a PostgreSQL database. It uses triggers to automatically record all changes applied to a database table in order to guarantee a complete audit trail of the contained data. It is centered around the idea that the database transaction is the natural auditing unit of data mutation. Any mutation on a participating table requires the developer first to record the operation metadata within the same database transaction. The metadata record is associated to the table change records by a constraint. On top of its database layer, Carbonite provides an API allowing developers to effortlessly retrieve, process, and purge the audit trails within the Elixir application. 🚧 This is work-in-progress and not yet ready to be used 🚧","ref":"readme.html","title":"README","type":"extras"},{"doc":"Convenient installation using migration functions Guaranteed consistency based on foreign-key constraints and ACID No mutation without recorded Change No Change without Transaction Customizable audit metadata (per transaction) Clear query-based interfaces to the audit trail Optional processing &amp; purging logic following the Outbox pattern Based on Ecto and Postgrex with no further dependencies. No configuration of PostgreSQL database needed","ref":"readme.html#features","title":"README - Features","type":"extras"},{"doc":"Requirements Due to its use of pg_current_xact_id , Carbonite requires ⚠️ PostgreSQL version 13 or above ⚠️. If you see an error message like the following, your PostgreSQL installation is too old: ** (Postgrex.Error) ERROR 42704 (undefined_object) type &quot;xid8&quot; does not exist Hex dependency # mix.exs def deps do [ { :carbonite , &quot;~&gt; 0.2.0&quot; } ] end","ref":"readme.html#installation","title":"README - Installation","type":"extras"},{"doc":"Carbonite implements the Change-Data-Capture pattern on top of PostgreSQL using database triggers. It keeps a central changes table where all mutations of participating tables are recorded. Each changes row is associated to a single row in the transactions table using PostgreSQL's internal transaction id as the foreign key. This leads to the following interesting properties: All changes created within a database transaction automatically and implicitly belong to the same record in the transactions table, even if they're created separatedly and agnostic of each other in the application logic. This gives the developer a &quot;natural&quot; way to group related changes into events (more on events later). As the changes table is associated to the transactions table via a non-nullable foreign key constraint, the entry in the transactions table must be created before any changes . Attempting to modify a versioned table without prior insertion into the transactions table will result in an error. The transactions table carries transactional metadata which can be set by the developer on creation. Consequently, much of Carbonite's logic lives in database functions and triggers. To get started, we need to create a migration using Ecto. ℹ️ Trigger vs. Write-Ahead-Log Existing solutions for CDC on top of a PostgreSQL database (e.g., Debezium ) often tail the &quot;Write-Ahead-Log&quot; instead of using database logic &amp; triggers to create change records. While this is likely more performant than using triggers, it makes it difficult to correlate changes on a transaction level as Carbonite does, and has different consistency guarantees. Carbonite makes this trade-off in favour of simplicity and universality. You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before.","ref":"readme.html#usage","title":"README - Usage","type":"extras"},{"doc":"The following migration installs Carbonite into its &quot;default prefix&quot;, a PostgreSQL schema aptly called carbonite_default , and installs the change capture trigger for an exemplary table called rabbits (in the public schema). In a real-world scenario, you will most likely want to install the trigger for a set of tables and optionally split the audit trail into multiple partitions. See Carbonite.Migrations for more information on migrations. mix ecto.gen.migration InstallCarbonite # priv/repo/migrations/20210704201534_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do use Ecto.Migration def up do # Creates carbonite_default schema and tables. Carbonite.Migrations . install_schema ( ) # For each table that you want to capture changes of, you need to install the trigger. Carbonite.Migrations . install_trigger ( :rabbits ) end def down do # Remove all triggers before dropping the schema. Carbonite.Migrations . drop_trigger ( :rabbits ) # Drop the schema &amp; tables. Carbonite.Migrations . drop_schema ( ) end end Primary Key Columns To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the table_pk column of the changes table. The table keeps an index on this column together with the table prefix and name. By default, Carbonite will try to copy the :id column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the primary_key_columns option of Carbonite.Migrations.install_trigger/2 and Carbonite.Migrations.configure_trigger/2 . # Disable PK copying Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ ] ) # Different name Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;identifier&quot; ] ) # Composite PK Carbonite.Migrations . install_trigger ( :rabbits , primary_key_columns : [ &quot;house&quot; , &quot;apartment_no&quot; ] ) Since the changes table keeps versions of a multitude of different source tables, primary keys are first cast to string (the table_pk column has type VARCHAR[] ). For composite primary keys, set the primary_key_columns option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array. Excluded and Filtered Columns In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the excluded_columns option. Excluded columns will not appear in the captured data. If an UPDATE on a table solely touches excluded columns, the entire UPDATE will not be recorded. Carbonite.Migrations . install_trigger ( :rabbits , excluded_columns : [ &quot;age&quot; ] ) If you still want to capture changes to a column (in the changed field), but don't need the exact data, you can make it a &quot;filtered&quot; column. These columns appear as &quot;[FILTERED]&quot; in the data field. ```elixir Carbonite.Migrations.install_trigger(:rabbits, filtered_columns: [&quot;age&quot;]) ``` If you forgot to exclude a column, you can reconfigure a trigger for a particular table using configure trigger/2 : ```elixir # in another migration Carbonite.Migrations.configure_trigger(:rabbits, excluded_columns: [&quot;age&quot;]) ``` #### Partitioning the Audit Trail Carbonite can install its tables into multiple database schemas using the prefix option. You can use this feature to &quot;partition&quot; your captured data. ```elixir Carbonite.Migrations.install_schema(carbonite_prefix: &quot;carbonite_lagomorpha&quot;) Carbonite.Migrations.install_trigger(:rabbits, carbonite_prefix: &quot;carbonite_lagomorpha&quot;) ``` If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage. ## Inserting a Transaction In your application logic, before modifying a versioned table like rabbits , you need to first create a Carbonite.Transaction record. ### With Ecto.Multi The easiest way to do so is using Carbonite.Multi.insert_transaction/2 within an Ecto.Multi operation: ```elixir Ecto.Multi.new() |&gt; Carbonite.Multi.insert_transaction(%{meta: %{type: &quot;rabbit_inserted&quot;}}) |&gt; Ecto.Multi.insert(:rabbit, &amp;MyApp.Rabbit.create_changeset(&amp;1.params)) |&gt; MyApp.Repo.transaction() ``` As you can see, the Carbonite.Transaction is a great place to store metadata for the operation. A &quot;transaction type&quot; would be an obvious choice to categorize the transactions. A user_id would be a good candidate for an transaction log, as well. ### Building a changeset for manual insertion If you don't have the luxury of an Ecto.Multi , you can create a changeset for a Carbonite.Transaction using Carbonite.Transaction.changeset/1 : ```elixir MyApp.Repo.transaction(fn -&gt; %{meta: %{type: &quot;rabbit_inserted&quot;}} |&gt; Carbonite.Transaction.changeset() |&gt; MyApp.Repo.insert!() # ... end) ``` ### Setting metadata outside of the Transaction In case you do not have access to metadata you want to persist in the Carbonite.Transaction at the code site where you create it, you can use Carbonite.Transaction.put_meta/2` to store metadata in the _process dictionary . This metadata is merged into the metadata given to Carbonite.Multi.insert_transaction/2 . elixir # e.g., in a controller or plug Carbonite.Transaction.put_meta(:user_id, ...) ## Retrieving data Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. Carbonite.Query offers a small suite of helper functions that make it easier to query the database for Transaction and Change records. ### Fetching Changes The Carbonite.Query.changes/2 function constructs an Ecto.Query from a schema struct, loading all changes stored for the given source record. elixir %MyApp.Rabbit{id: 1} |&gt; Carbonite.Query.changes() |&gt; MyApp.Repo.all() ## Testing / Bypassing Carbonite One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing Carbonite.Transaction beforehand (due to the foreign key constraint between changes and transactions ). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. To bypass the capture trigger, Carbonite's trigger configuration provides a toggle mechanism consisting of two fields: mode and override_transaction_id . The former you set while installing the trigger on a table in a migration, while the latter allows to &quot;override&quot; whatever has been set at runtime, and only for the current transaction. If you are using Ecto's SQL sandbox for running transactional tests, this means the override is going to be active until the end of the test case. As a result, you have two options: 1. Leave the mode at the default value of :capture and turn off capturing as needed by switching to &quot;override mode&quot;. This means for every test case where you do not care about change capturing, you explicitly disable the trigger before any database calls; for instance, in a ExUnit setup block. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a Carbonite.Transaction . It is, however, still pretty expensive at ~1 additional SQL call per test case. 2. Set the mode to :ignore on all triggers in your :test environment and instead selectively turn on capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well. The following code snippet illustrates the second approach: elixir # config/config.exs config :my_app, carbonite_mode: :capture # config/test.exs config :my_app, carbonite_mode: :ignore # priv/repo/migrations/000000000000_install_carbonite.exs defmodule MyApp.Repo.Migrations.InstallCarbonite do @mode Application.compile_env!(:my_app, :carbonite_mode) def up do # ... Carbonite.Migrations.install_trigger(:rabbits, mode: @mode) end end # test/support/carbonite_helpers.exs defmodule MyApp.CarboniteHelpers do def carbonite_override_mode(_) do Ecto.Multi.new() |&gt; Carbonite.Multi.override_mode() |&gt; MyApp.Repo.transaction() :ok end end # test/some_test.exs def SomeTest do use MyApp.DataCase import MyApp.CarboniteHelpers setup [:carbonite_override_mode] end &lt;!-- MDOC --&gt; ## Acknowledgements ### Inspiration The trigger-based table versioning derives from audit_trigger_91plus , an &quot;example of a generic trigger function&quot; hosted in the PostgreSQL wiki. ### Artwork The amazing Carbonite logo has been designed by Petra Herberger . Thank you! 💜 ## Copyright and License Copyright © 2021, Bitcrowd GmbH. You may use and redistribute Carbonite and its source code under the Apache 2.0 License, see the LICENSE file for details.","ref":"readme.html#installing-the-schema-triggers","title":"README - Installing the schema &amp; triggers","type":"extras"},{"doc":"","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Added Store primary key(s) on changes table and index them Add Carbonite.Query module current_transaction/2 allows to fetch the ongoing transaction (for sandbox tests) changes/2 allows to fetch the changes of an invidual source record Update Postgrex to 0.15.11 and drop local Xid8 type Add mode field to trigger (capture or ignore) Add &quot;override mode&quot; reversing the mode option for the current transaction to enable/disable capturing on demand (e.g. in tests) Add filtered columns Changed Moved top-level functions to nested modules Transaction and Multi Made table_pk be NULL when primary_key_columns is an empty array Default primary_key_columns to [&quot;id&quot;] Renamed prefix option to carbonite_prefix on install_schema/2 for consistency","ref":"changelog.html#unreleased","title":"Changelog - Unreleased","type":"extras"},{"doc":"Initial release.","ref":"changelog.html#0-1-0-2021-09-01","title":"Changelog - [0.1.0] - 2021-09-01","type":"extras"}]